/*********************************************************************/
/* queues.h                                                          */
/* typedefs, macros and inline functs for various queues management  */
/* see COPYRIGHT in parent directory                                 */
/*********************************************************************/

#ifndef NEBLO_QUEUES_H
#define NEBLO_QUEUES_H


#include <neblo_.h>

#undef DEBUG_Q                /* dealing with Queues (of many kinds) */


/*
 * queue of announced (expected) friendship requests.
 */

typedef struct expd_reqst
{
  struct in_addr foreign_addr;   /* IP addr of requestor                  */
  u_int16_t foreign_port;        /* listen port of requestor              */
  unsigned int pred_succ;        /* 1 if expected new predecessor,        */
                                 /* 2 if expected new successor,          */
                                 /* 0 otherwise.                          */
  struct expd_reqst *next;
  RSA *pubkey;                   /* pubkey of requestor */
  struct timeval timestamp;      /* time of insert, for the garbage coll. */
} expd_reqst_t;

void init_expd_reqst_q ( void );

RSA  *search_expd_reqst ( struct in_addr, u_int16_t, unsigned int, pthread_t );

int   insert_expd_reqst
        ( struct in_addr, u_int16_t, unsigned int, RSA *, pthread_t );

void *clean_expd_reqsts ( void * );


/*
 * queue of pending read requests:
 * request of kind "read" issued to a remote peer but not yet answered
 */

#define I_MAX_P_READ 8192       /* initial room for pending "read" reqsts */
#define MAX_P_READ  65536       /* max no. of pending "read" reqsts */

typedef struct pend_read_req
{
  pthread_mutex_t lock;         /* lock on entire struct */
  unsigned int in_use;          /* how many pend. reqs refer to this handle */
  unsigned int done;            /* significant on originator peer only */
                                /* >0 iff request served (data available) */
                                /* 0 iff no answer arrived yet */
  neblo_rhandle handle;         /* id. for this req */
  peer_t *from;                 /* req came from this peer, so response */
                                /* will be eventually backwarded to it. */
                                /* NULL iff req comes from local peer */
  peer_t *to;                   /* req has been forwarded to this peer, so */
                                /* we expect a response from it. */
  neblo_rhandle from_rhandle;   /* id. of request at the "from" peer */
                                /* R_NONE iff req comes from local peer */
  neblo_port port;              /* read through this NEBLO port of recipient */
  neblo_addr addr;              /* dest. of the read req in the 64-bit space */
  neblo_addr real_addr;         /* read at this addr, in the 64-bit space */
  void *data[NEBLO_MAXCP];      /* responses will be stored at these addrs */
                                /* NULL on non-originator peers */
  unsigned int maxlen;          /* max data size for this read req */
                                /* 0 (meaningless) on non-originator peers */
  pthread_mutex_t wait_sem;     /* semaphore associated to condition below */
  pthread_cond_t wait_cond;     /* condition; will unblock when "read" done */
  struct timeval timestamp;     /* time of insertion, for the garbage coll. */
} pend_read_t;

extern pend_read_t       *pend_read;
extern unsigned int       pend_read_head;
extern unsigned int       pend_read_tail;
extern unsigned int       pend_read_max;
extern pthread_mutex_t    p_read_tail_lock;

void  init_pend_read_q ( void );

int   insert_pend_read
        ( neblo_port, neblo_addr, void *, unsigned int, peer_t *,
        neblo_rhandle, peer_t *, neblo_rhandle * );

pend_read_t *
      search_pend_read ( neblo_rhandle );

void  cancel_pend_read ( pend_read_t * );

void  decrease_pend_read ( pend_read_t * );

void *clean_pend_reads ( void * );

void  clean_pend_reads_peer ( peer_t * );

int   pend_read_ref ( peer_t *f );


/*
 * queue of read/write/data msgs that could not be forwarded yet, and thus
 * must be retried
 */

#define FIRST_ATTMPT 0
#define RETRY 1
#define LAST_CHANCE 2

typedef struct retry_reqst
{
  rwd_t kind;                     /* READ, WRITE, DATA, REFR */
  reqst_t *reqst;                 /* the request to be retried */
  unsigned int pending;           /* 1 if still to be served, 0 otherwise */
  struct retry_reqst *next;
  struct timeval timestamp;       /* time of insertion, for the garbg coll */
} retry_reqst_t;

void  init_retry_queue ( void );

int   insert_retry_queue ( rwd_t, reqst_t *, pthread_t );

void *clean_retry_queue ( void * );


#endif
